import struct


def testBit(int_type: int, offset: int) -> int:
    """
    testBit() returns a nonzero result, 2**offset, if the bit at
    'offset' is one.
    """
    mask = 1 << offset
    return int_type & mask


def setBit(int_type: int, offset: int) -> int:
    """setBit() returns an integer with the bit at 'offset' set to 1."""
    mask = 1 << offset
    return int_type | mask


def clearBit(int_type: int, offset: int) -> int:
    """clearBit() returns an integer with the bit at 'offset' cleared."""
    mask = ~(1 << offset)
    return int_type & mask


def toggleBit(int_type: int, offset: int) -> int:
    """
    toggleBit() returns an integer with the bit at
    'offset' inverted, 0 -> 1 and 1 -> 0.
    """
    mask = 1 << offset
    return int_type ^ mask


def bytes_to_int(bytes: bytes, signed: bool = False) -> int:
    return int.from_bytes(bytes, "little", signed=signed)


def bytes_to_float(bytes: bytes) -> float:
    return float(struct.unpack("<d", bytes)[0])



# Helper script for tests which compares content of two pyc files

import marshal
import sys

from marshalparser.magic import get_pyc_header_lenght

PYC = bool(int(sys.argv[1]))
original_filename = sys.argv[2]
fixed_filename = sys.argv[3]

if PYC:
    pyc_header_len = get_pyc_header_lenght(sys.version_info[:2])

with open(original_filename, mode="rb") as fh:
    if PYC:
        fh.seek(pyc_header_len)
    original = marshal.load(fh)

with open(fixed_filename, mode="rb") as fh:
    if PYC:
        fh.seek(pyc_header_len)
    fixed = marshal.load(fh)

assert original == fixed



import struct
from functools import lru_cache
from typing import Optional, Tuple


def inclusive_range(f: int, t: int) -> range:
    """Returns range including both ends"""
    return range(f, t + 1)


# Based on CPython source code
# Lib/importlib/_bootstrap_external.py
MAGIC_NUMBERS_RANGES = (
    (inclusive_range(3360, 3361), (3, 6)),
    (inclusive_range(3370, 3379), (3, 6)),
    (inclusive_range(3390, 3394), (3, 7)),
    (inclusive_range(3400, 3401), (3, 8)),
    (inclusive_range(3410, 3413), (3, 8)),
    (inclusive_range(3420, 3425), (3, 9)),
    (inclusive_range(3430, 3439), (3, 10)),
    (inclusive_range(3450, 3495), (3, 11)),
    # These ranges have to be adjusted when the final releases are out
    (inclusive_range(3500, 3549), (3, 12)),
    (inclusive_range(3550, 3599), (3, 13)),
)


@lru_cache()
def python_version_from_magic_number(number: int) -> Optional[Tuple[int, int]]:
    for numbers_range, version in MAGIC_NUMBERS_RANGES:
        if number in numbers_range:
            return version
    return None


def get_pyc_python_version(
    *, filename: Optional[str] = None, bytes: Optional[bytes] = None
) -> Optional[Tuple[int, int]]:
    """Get the version of Python from pyc header (magic number)
    or None if it cannot be detected. Takes a filename to read
    or first 4 bytes from it."""
    magic = None
    if filename:
        with open(filename, "rb") as file:
            magic = file.read(4)
    elif bytes:
        magic = bytes

    if not magic:
        raise RuntimeError("Either filename or bytes has to be specified!")

    magic_data = struct.unpack("<H2B", magic)
    python_version = python_version_from_magic_number(magic_data[0])

    return python_version


def get_pyc_header_lenght(python_version: Tuple[int, int]) -> int:
    """Returns pyc header lenght (number of bytes) for Python version"""
    if python_version >= (3, 7):
        return 16
    elif python_version >= (3, 3):
        return 12
    else:
        return 8



import filecmp
import os
import sys
from pathlib import Path
from subprocess import STDOUT, CalledProcessError, check_call, check_output

import pytest

from marshalparser.magic import get_pyc_python_version

PYTHON_VERSION = "{}.{}".format(*sys.version_info[:2])
CMD = ["python" + PYTHON_VERSION, "-m", "marshalparser", "-f"]


def fixed_filename(original_filename):
    original_filename = Path(original_filename)
    return original_filename.with_suffix(".fixed" + original_filename.suffix)


def param(p, *, skip_reason=None):
    """Return a pytest.param with stringified id, for nicer verbose output

    When skip_reson is set, the pytest.param will be marked as skipped."""
    if skip_reason is None:
        return pytest.param(p, id=str(p))
    return pytest.param(p,
                        id=str(p),
                        marks=[pytest.mark.skip(reason=skip_reason)])


def generate_test_data():
    for python_version_dir in (Path("test") / "python_stdlib").glob("*"):
        python_version = python_version_dir.name
        try:
            check_call(("python" + python_version, "-c", "pass"))
        except FileNotFoundError:
            skip_reason = f"python{python_version} not found"
        else:
            skip_reason = None
        yield from (
            param(p, skip_reason=skip_reason)
            for p in python_version_dir.glob("*.pyc")
        )
    yield from (param(p) for p in (Path("test") / "pure_marshal").glob("*"))
    yield from (param(p) for p in (Path("test") / "renamed_pycs").glob("*"))


@pytest.mark.parametrize("original_filename", generate_test_data())
def test_complete(original_filename, tmp_path):
    # To be able to run tests in parallel, we create
    # a symlink for each test file so the fixed file appears
    # next to the symlink in the temp dir instead of the
    # test folder next to the original files.
    filename = tmp_path / original_filename.name
    filename.symlink_to(original_filename.absolute())

    # This command uses the Python we are testing with
    # because for example we can run marshalparser with
    # Python 3.9 and fix pyc file for Python 3.6
    output = check_output(CMD + [filename], encoding="utf-8")

    assert os.path.isfile(fixed_filename(filename))

    if filecmp.cmp(filename, fixed_filename(filename), shallow=False):
        pytest.skip("Fixed file is the same as original, nothing to check")

    assert output.endswith(
        f" unused FLAG_REFs from {fixed_filename(filename)}\n"
    )
    assert output[:8] == "Removed "

    # To compare two pyc files, we need to use Python
    # they were compiled by
    python_version = get_pyc_python_version(filename=filename)
    if python_version:
        python_version_str = "{}.{}".format(*python_version)
        # pass this to marshal_content_check.py so we don't have to
        # check the header there again
        check_pyc = True
    else:
        # For non-pyc files, we use the current Python for content check
        python_version_str = PYTHON_VERSION
        check_pyc = False

    CHECK_CMD = [
        "python" + python_version_str,
        "marshal_content_check.py",
        str(int(check_pyc)),
        filename,
        fixed_filename(filename),
    ]

    try:
        check_call(CHECK_CMD)
    except FileNotFoundError:
        # this could still happen with test/renamed_pycs
        pytest.skip(
            f"python{python_version_str} not found! Cannot check the result."
        )

    # Fixing the already fixed file should make no changes
    again_output = check_output(
        CMD + ["-o", fixed_filename(filename)], encoding="utf-8"
    )
    assert (
        again_output == f"No unused FLAG_REFs in {fixed_filename(filename)}\n"
    )


test_files = sorted((Path("test") / "renamed_pycs").glob("*"))
three_doubles = [(test_files[n], test_files[n + 1]) for n in range(0, 6, 2)]


@pytest.mark.parametrize("original_filenames", three_doubles)
def test_run_with_more_than_one_file(original_filenames, tmp_path):
    filenames = []
    for original_filename in original_filenames:
        filename = tmp_path / original_filename.name
        filename.symlink_to(original_filename.absolute())
        filenames.append(filename)
    check_call(CMD + [*filenames])


def test_empty_file(tmp_path):
    filename = tmp_path / "empty_file.pyc"
    filename.touch()
    with pytest.raises(CalledProcessError) as e:
        check_output(CMD + [filename], encoding="utf-8", stderr=STDOUT)
    assert e.value.returncode == 1
    assert f"File {filename} is empty!" in e.value.output



import marshal

data = {
    "integer": 42,
    "longs": (25487958525413698547, 1 << 64),
    "float": 42.88,
    "string": "hello world",
    "list_of_simple_objects": [True, False, None, ...],
    "set": set((9, 8, 7)),
    "tuple": (6, 5, 4),
    "dict": dict(a=1, b=2, c=3),
    "complex": complex(15, 2),
    "frozenset": frozenset((99, 999, 9999)),
    "nested_tuples": (),
}

for i in range(25):
    data["nested_tuples"] = (data["nested_tuples"], data["nested_tuples"])

for key, value in data.items():
    with open(f"test/pure_marshal/{key}.dat", mode="wb") as fh:
        marshal.dump(value, fh)
